% Log-barrier constraint function
%
% constraints is a cell array of constraint functions
% xx is the current iterate
%
% value is the function value at xx
% gradient is the gradient at xx
% hessian is the hessian at xx
%
function [ value, gradient, hessian ] = objective_constraints( constraints, xx )

	if ( length( constraints ) < 1 ); error( 'constraint cell array cannot be empty' ); end

	% we need to make sure that we get infinity (not a complex number) for xx which
	% violate the constraints
	function ll = safelog( pp )
		ll = log( max( 0, pp ) );
	end


	values = {};
	gradients = {};
	hessians = {};

	% fill in values{}, gradients{} and hessians{} with the values, gradients and
	% hessians of each constraint
	for ii = 1:length( constraints )

		% find the value, matrix derivatives, and matrix second derivatives of the
		% constraint
		if ( nargout < 2 )
			[ constraint_value ] = constraints{ ii }( xx );
		elseif ( nargout < 3 )
			[ constraint_value, constraint_gradient ] = constraints{ ii }( xx );
			% at this point, constraint_gradient( ii, :, : ) contains the matrix
			% derivative of the constraint w.r.t. xx_ii. we permute the indices to
			% make constraint_gradient( :, :, ii ) this derivative, ensuring that
			% when we take the necessary array slices, the result is a matrix
			constraint_gradient = permute( constraint_gradient, [ 2, 3, 1 ] );
		else
			[ constraint_value, constraint_gradient, constraint_hessian ] = constraints{ ii }( xx );
			% at this point, constraint_gradient( ii, :, : ) contains the matrix
			% derivative of the constraint w.r.t. xx_ii, and constraint_hessian(
			% ii, jj, :, : ) is the second derivative w.r.t. xx_ii*xx_jj. we
			% permute the indices to make constraint_gradient( :, :, ii ) and
			% constraint_hessian( :, :, ii, jj ) this derivative and Hessian,
			% ensuring that when we take the necessary array slices, the result is
			% a matrix
			constraint_gradient = permute( constraint_gradient, [ 2, 3, 1 ] );
			constraint_hessian = permute( constraint_hessian, [ 3, 4, 1, 2 ] );
		end

		value = -sum( safelog( eig( constraint_value ) ) );
		values = [ values, { value } ];

		if ( nargout > 1 )

			gradient = zeros( size( constraint_gradient, 3 ), 1 );
			for jj = 1:length( constraint_gradient )
				%gradient( jj ) = ( TODO: derivative of -logdet( constraint ) with respect to xx( jj ), evaluated at xx );
				%    NOTE: constraint_value is the value of the constraint function at xx
				%    NOTE: constraint_gradient( :, :, ii ) is the matrix derivative of the constraint function with respect to xx( ii ), evaluated at xx
			end
			gradients = [ gradients, { gradient } ];

			if ( nargout > 2 )

				hessian = zeros( size( constraint_hessian, 3 ), size( constraint_hessian, 4 ) );
				for jj = 1:size( constraint_hessian, 3 )
					for kk = 1:size( constraint_hessian, 4 )
						%hessian( jj, kk ) = ( TODO: second derivative of -logdet( constraint ) with respect to xx( jj )xx( kk ), evaluated at xx );
						%    NOTE: constraint_value is the value of the constraint function at xx
						%    NOTE: constraint_gradient( :, :, ii ) is the matrix derivative of the constraint function with respect to xx( ii ), evaluated at xx
						%    NOTE: constraint_hessian( :, :, ii, jj ) is the matrix second derivative of the constraint function with respect to xx( ii )xx( jj ), evaluated at xx
					end
				end
				hessians = [ hessians, { hessian } ];

			end
		end

	end


	% sum the values for all constraints
	value = values{ 1 };
	for ii = 2:length( values )
		value = value + values{ ii };
	end

	if ( nargout > 1 )

		% sum the gradients for all constraints
		gradient = gradients{ 1 };
		for ii = 2:length( gradients )
			gradient = gradient + gradients{ ii };
		end

		if ( nargout > 2 )

			% sum the hessians for all constraints
			hessian = hessians{ 1 };
			for ii = 2:length( hessians )
				hessian = hessian + hessians{ ii };
			end
		end
	end

end
